// https://github.com/d3/d3-delaunay Version 3.1.2. Copyright 2018 Observable, Inc.
// https://github.com/mapbox/delaunator Version 2.0.0. Copyright 2017, Mapbox, Inc.
(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
    typeof define === 'function' && define.amd ? define(['exports'], factory) :
    (factory((global.d3 = global.d3 || {})));
}(this, (function (exports) { 'use strict';

    Delaunator.from = function (points, getX, getY) {
        if (!getX) getX = defaultGetX;
        if (!getY) getY = defaultGetY;

        var n = points.length;
        var coords = new Float64Array(n * 2);

        for (var i = 0; i < n; i++) {
            var p = points[i];
            coords[2 * i] = getX(p);
            coords[2 * i + 1] = getY(p);
        }

        return new Delaunator(coords);
    };

    function Delaunator(coords) {
        if (!ArrayBuffer.isView(coords)) throw new Error('Expected coords to be a typed array.');

        var minX = Infinity;
        var minY = Infinity;
        var maxX = -Infinity;
        var maxY = -Infinity;

        var n = coords.length >> 1;
        var ids = this.ids = new Uint32Array(n);

        this.coords = coords;

        for (var i = 0; i < n; i++) {
            var x = coords[2 * i];
            var y = coords[2 * i + 1];
            if (x < minX) minX = x;
            if (y < minY) minY = y;
            if (x > maxX) maxX = x;
            if (y > maxY) maxY = y;
            ids[i] = i;
        }

        var cx = (minX + maxX) / 2;
        var cy = (minY + maxY) / 2;

        var minDist = Infinity;
        var i0, i1, i2;

        // pick a seed point close to the centroid
        for (i = 0; i < n; i++) {
            var d = dist(cx, cy, coords[2 * i], coords[2 * i + 1]);
            if (d < minDist) {
                i0 = i;
                minDist = d;
            }
        }

        minDist = Infinity;

        // find the point closest to the seed
        for (i = 0; i < n; i++) {
            if (i === i0) continue;
            d = dist(coords[2 * i0], coords[2 * i0 + 1], coords[2 * i], coords[2 * i + 1]);
            if (d < minDist && d > 0) {
                i1 = i;
                minDist = d;
            }
        }

        var minRadius = Infinity;

        // find the third point which forms the smallest circumcircle with the first two
        for (i = 0; i < n; i++) {
            if (i === i0 || i === i1) continue;

            var r = circumradius(
                coords[2 * i0], coords[2 * i0 + 1],
                coords[2 * i1], coords[2 * i1 + 1],
                coords[2 * i], coords[2 * i + 1]);

            if (r < minRadius) {
                i2 = i;
                minRadius = r;
            }
        }

        if (minRadius === Infinity) {
            throw new Error('No Delaunay triangulation exists for this input.');
        }

        // swap the order of the seed points for counter-clockwise orientation
        if (area(coords[2 * i0], coords[2 * i0 + 1],
            coords[2 * i1], coords[2 * i1 + 1],
            coords[2 * i2], coords[2 * i2 + 1]) < 0) {

            var tmp = i1;
            i1 = i2;
            i2 = tmp;
        }

        var i0x = coords[2 * i0];
        var i0y = coords[2 * i0 + 1];
        var i1x = coords[2 * i1];
        var i1y = coords[2 * i1 + 1];
        var i2x = coords[2 * i2];
        var i2y = coords[2 * i2 + 1];

        var center = circumcenter(i0x, i0y, i1x, i1y, i2x, i2y);
        this._cx = center.x;
        this._cy = center.y;

        // sort the points by distance from the seed triangle circumcenter
        quicksort(ids, coords, 0, ids.length - 1, center.x, center.y);

        // initialize a hash table for storing edges of the advancing convex hull
        this._hashSize = Math.ceil(Math.sqrt(n));
        this._hash = [];
        for (i = 0; i < this._hashSize; i++) this._hash[i] = null;

        // initialize a circular doubly-linked list that will hold an advancing convex hull
        var e = this.hull = insertNode(coords, i0);
        this._hashEdge(e);
        e.t = 0;
        e = insertNode(coords, i1, e);
        this._hashEdge(e);
        e.t = 1;
        e = insertNode(coords, i2, e);
        this._hashEdge(e);
        e.t = 2;

        var maxTriangles = 2 * n - 5;
        var triangles = this.triangles = new Uint32Array(maxTriangles * 3);
        var halfedges = this.halfedges = new Int32Array(maxTriangles * 3);

        this.trianglesLen = 0;

        this._addTriangle(i0, i1, i2, -1, -1, -1);

        var xp, yp;
        for (var k = 0; k < ids.length; k++) {
            i = ids[k];
            x = coords[2 * i];
            y = coords[2 * i + 1];

            // skip duplicate points
            if (x === xp && y === yp) continue;
            xp = x;
            yp = y;

            // skip seed triangle points
            if ((x === i0x && y === i0y) ||
                (x === i1x && y === i1y) ||
                (x === i2x && y === i2y)) continue;

            // find a visible edge on the convex hull using edge hash
            var startKey = this._hashKey(x, y);
            var key = startKey;
            var start;
            do {
                start = this._hash[key];
                key = (key + 1) % this._hashSize;
            } while ((!start || start.removed) && key !== startKey);

            e = start;
            while (area(x, y, e.x, e.y, e.next.x, e.next.y) >= 0) {
                e = e.next;
                if (e === start) {
                    throw new Error('Something is wrong with the input points.');
                }
            }

            var walkBack = e === start;

            // add the first triangle from the point
            var t = this._addTriangle(e.i, i, e.next.i, -1, -1, e.t);

            e.t = t; // keep track of boundary triangles on the hull
            e = insertNode(coords, i, e);

            // recursively flip triangles from the point until they satisfy the Delaunay condition
            e.t = this._legalize(t + 2);
            if (e.prev.prev.t === halfedges[t + 1]) {
                e.prev.prev.t = t + 2;
            }

            // walk forward through the hull, adding more triangles and flipping recursively
            var q = e.next;
            while (area(x, y, q.x, q.y, q.next.x, q.next.y) < 0) {
                t = this._addTriangle(q.i, i, q.next.i, q.prev.t, -1, q.t);
                q.prev.t = this._legalize(t + 2);
                this.hull = removeNode(q);
                q = q.next;
            }

            if (walkBack) {
                // walk backward from the other side, adding more triangles and flipping
                q = e.prev;
                while (area(x, y, q.prev.x, q.prev.y, q.x, q.y) < 0) {
                    t = this._addTriangle(q.prev.i, i, q.i, -1, q.t, q.prev.t);
                    this._legalize(t + 2);
                    q.prev.t = t;
                    this.hull = removeNode(q);
                    q = q.prev;
                }
            }

            // save the two new edges in the hash table
            this._hashEdge(e);
            this._hashEdge(e.prev);
        }

        // trim typed triangle mesh arrays
        this.triangles = triangles.subarray(0, this.trianglesLen);
        this.halfedges = halfedges.subarray(0, this.trianglesLen);
    }

    Delaunator.prototype = {

        _hashEdge: function (e) {
            this._hash[this._hashKey(e.x, e.y)] = e;
        },

        _hashKey: function (x, y) {
            var dx = x - this._cx;
            var dy = y - this._cy;
            // use pseudo-angle: a measure that monotonically increases
            // with real angle, but doesn't require expensive trigonometry
            var p = 1 - dx / (Math.abs(dx) + Math.abs(dy));
            return Math.floor((2 + (dy < 0 ? -p : p)) / 4 * this._hashSize);
        },

        _legalize: function (a) {
            var triangles = this.triangles;
            var coords = this.coords;
            var halfedges = this.halfedges;

            var b = halfedges[a];

            var a0 = a - a % 3;
            var b0 = b - b % 3;

            var al = a0 + (a + 1) % 3;
            var ar = a0 + (a + 2) % 3;
            var bl = b0 + (b + 2) % 3;

            var p0 = triangles[ar];
            var pr = triangles[a];
            var pl = triangles[al];
            var p1 = triangles[bl];

            var illegal = inCircle(
                coords[2 * p0], coords[2 * p0 + 1],
                coords[2 * pr], coords[2 * pr + 1],
                coords[2 * pl], coords[2 * pl + 1],
                coords[2 * p1], coords[2 * p1 + 1]);

            if (illegal) {
                triangles[a] = p1;
                triangles[b] = p0;

                this._link(a, halfedges[bl]);
                this._link(b, halfedges[ar]);
                this._link(ar, bl);

                var br = b0 + (b + 1) % 3;

                this._legalize(a);
                return this._legalize(br);
            }

            return ar;
        },

        _link: function (a, b) {
            this.halfedges[a] = b;
            if (b !== -1) this.halfedges[b] = a;
        },

        // add a new triangle given vertex indices and adjacent half-edge ids
        _addTriangle: function (i0, i1, i2, a, b, c) {
            var t = this.trianglesLen;

            this.triangles[t] = i0;
            this.triangles[t + 1] = i1;
            this.triangles[t + 2] = i2;

            this._link(t, a);
            this._link(t + 1, b);
            this._link(t + 2, c);

            this.trianglesLen += 3;

            return t;
        }
    };

    function dist(ax, ay, bx, by) {
        var dx = ax - bx;
        var dy = ay - by;
        return dx * dx + dy * dy;
    }

    function area(px, py, qx, qy, rx, ry) {
        return (qy - py) * (rx - qx) - (qx - px) * (ry - qy);
    }

    function inCircle(ax, ay, bx, by, cx, cy, px, py) {
        ax -= px;
        ay -= py;
        bx -= px;
        by -= py;
        cx -= px;
        cy -= py;

        var ap = ax * ax + ay * ay;
        var bp = bx * bx + by * by;
        var cp = cx * cx + cy * cy;

        return ax * (by * cp - bp * cy) -
               ay * (bx * cp - bp * cx) +
               ap * (bx * cy - by * cx) < 0;
    }

    function circumradius(ax, ay, bx, by, cx, cy) {
        bx -= ax;
        by -= ay;
        cx -= ax;
        cy -= ay;

        var bl = bx * bx + by * by;
        var cl = cx * cx + cy * cy;

        if (bl === 0 || cl === 0) return Infinity;

        var d = bx * cy - by * cx;
        if (d === 0) return Infinity;

        var x = (cy * bl - by * cl) * 0.5 / d;
        var y = (bx * cl - cx * bl) * 0.5 / d;

        return x * x + y * y;
    }

    function circumcenter(ax, ay, bx, by, cx, cy) {
        bx -= ax;
        by -= ay;
        cx -= ax;
        cy -= ay;

        var bl = bx * bx + by * by;
        var cl = cx * cx + cy * cy;

        var d = bx * cy - by * cx;

        var x = (cy * bl - by * cl) * 0.5 / d;
        var y = (bx * cl - cx * bl) * 0.5 / d;

        return {
            x: ax + x,
            y: ay + y
        };
    }

    // create a new node in a doubly linked list
    function insertNode(coords, i, prev) {
        var node = {
            i: i,
            x: coords[2 * i],
            y: coords[2 * i + 1],
            t: 0,
            prev: null,
            next: null,
            removed: false
        };

        if (!prev) {
            node.prev = node;
            node.next = node;

        } else {
            node.next = prev.next;
            node.prev = prev;
            prev.next.prev = node;
            prev.next = node;
        }
        return node;
    }

    function removeNode(node) {
        node.prev.next = node.next;
        node.next.prev = node.prev;
        node.removed = true;
        return node.prev;
    }

    function quicksort(ids, coords, left, right, cx, cy) {
        var i, j, temp;

        if (right - left <= 20) {
            for (i = left + 1; i <= right; i++) {
                temp = ids[i];
                j = i - 1;
                while (j >= left && compare(coords, ids[j], temp, cx, cy) > 0) ids[j + 1] = ids[j--];
                ids[j + 1] = temp;
            }
        } else {
            var median = (left + right) >> 1;
            i = left + 1;
            j = right;
            swap(ids, median, i);
            if (compare(coords, ids[left], ids[right], cx, cy) > 0) swap(ids, left, right);
            if (compare(coords, ids[i], ids[right], cx, cy) > 0) swap(ids, i, right);
            if (compare(coords, ids[left], ids[i], cx, cy) > 0) swap(ids, left, i);

            temp = ids[i];
            while (true) {
                do i++; while (compare(coords, ids[i], temp, cx, cy) < 0);
                do j--; while (compare(coords, ids[j], temp, cx, cy) > 0);
                if (j < i) break;
                swap(ids, i, j);
            }
            ids[left + 1] = ids[j];
            ids[j] = temp;

            if (right - i + 1 >= j - left) {
                quicksort(ids, coords, i, right, cx, cy);
                quicksort(ids, coords, left, j - 1, cx, cy);
            } else {
                quicksort(ids, coords, left, j - 1, cx, cy);
                quicksort(ids, coords, i, right, cx, cy);
            }
        }
    }

    function compare(coords, i, j, cx, cy) {
        var d1 = dist(coords[2 * i], coords[2 * i + 1], cx, cy);
        var d2 = dist(coords[2 * j], coords[2 * j + 1], cx, cy);
        return (d1 - d2) || (coords[2 * i] - coords[2 * j]) || (coords[2 * i + 1] - coords[2 * j + 1]);
    }

    function swap(arr, i, j) {
        var tmp = arr[i];
        arr[i] = arr[j];
        arr[j] = tmp;
    }

    function defaultGetX(p) {
        return p[0];
    }
    function defaultGetY(p) {
        return p[1];
    }

    class Path {
      constructor() {
        this._x0 = this._y0 = // start of current subpath
        this._x1 = this._y1 = null; // end of current subpath
        this._ = "";
      }
      moveTo(x, y) {
        this._ += `M${this._x0 = this._x1 = +x},${this._y0 = this._y1 = +y}`;
      }
      closePath() {
        if (this._x1 !== null) {
          this._x1 = this._x0, this._y1 = this._y0;
          this._ += "Z";
        }
      }
      lineTo(x, y) {
        this._ += `L${this._x1 = +x},${this._y1 = +y}`;
      }
      arc(x, y, r) {
        x = +x, y = +y, r = +r;
        const x0 = x + r;
        const y0 = y;
        if (r < 0) throw new Error("negative radius");
        if (this._x1 === null) this._ += `M${x0},${y0}`;
        else if (Math.abs(this._x1 - x0) > epsilon || Math.abs(this._y1 - y0) > epsilon) this._ += "L" + x0 + "," + y0;
        if (!r) return;
        this._ += `A${r},${r},0,1,1,${x - r},${y}A${r},${r},0,1,1,${this._x1 = x0},${this._y1 = y0}`;
      }
      rect(x, y, w, h) {
        this._ += `M${this._x0 = this._x1 = +x},${this._y0 = this._y1 = +y}h${+w}v${+h}h${-w}Z`;
      }
      value() {
        return this._ || null;
      }
    }

    class Polygon {
      constructor() {
        this._ = [];
      }
      moveTo(x, y) {
        this._.push([x, y]);
      }
      closePath() {
        this._.push(this._[0].slice());
      }
      lineTo(x, y) {
        this._.push([x, y]);
      }
      value() {
        return this._;
      }
    }

    class Voronoi {
      constructor(delaunay, [xmin, ymin, xmax, ymax] = [0, 0, 960, 500]) {
        if (!((xmax = +xmax) >= (xmin = +xmin)) || !((ymax = +ymax) >= (ymin = +ymin))) throw new Error("invalid bounds");
        const {points, halfedges, hull, triangles} = this.delaunay = delaunay;
        const circumcenters = this.circumcenters = new Float64Array(triangles.length / 3 * 2);
        const edges = this.edges = new Uint32Array(halfedges.length);
        const index = this.index = new Uint32Array(points.length);
        const vectors = this.vectors = new Float64Array(points.length * 2);
        this.xmax = xmax, this.xmin = xmin;
        this.ymax = ymax, this.ymin = ymin;

        // Compute cell topology.
        for (let i = 0, e = 0, m = halfedges.length; i < m; ++i) {
          const t = triangles[i]; // Cell vertex.
          if (index[t * 2] !== index[t * 2 + 1]) continue; // Already connected.
          const e0 = index[t * 2] = e;
          let j = i;

          do { // Walk forward.
            edges[e++] = Math.floor(j / 3);
            j = halfedges[j];
            if (j === -1) break; // Went off the convex hull.
            j = j % 3 === 2 ? j - 2 : j + 1;
            if (triangles[j] !== t) break; // Bad triangulation; break early.
          } while (j !== i);

          if (j !== i) { // Stopped when walking forward; walk backward.
            const k = halfedges[i % 3 === 0 ? i + 2 : i - 1];
            const e1 = e;
            j = i;
            do {
              j = halfedges[j % 3 === 0 ? j + 2 : j - 1];
              if (j === -1 || triangles[j] !== t) break;
              edges[e++] = Math.floor(j / 3);
            } while (j !== k);
            if (e1 < e) {
              edges.subarray(e0, e1).reverse();
              edges.subarray(e0, e).reverse();
            }
          }

          index[t * 2 + 1] = e;
        }

        // Compute circumcenters.
        for (let i = 0, j = 0, n = triangles.length; i < n; i += 3, j += 2) {
          const t1 = triangles[i] * 2;
          const t2 = triangles[i + 1] * 2;
          const t3 = triangles[i + 2] * 2;
          const x1 = points[t1];
          const y1 = points[t1 + 1];
          const x2 = points[t2];
          const y2 = points[t2 + 1];
          const x3 = points[t3];
          const y3 = points[t3 + 1];
          const a2 = x1 - x2;
          const a3 = x1 - x3;
          const b2 = y1 - y2;
          const b3 = y1 - y3;
          const d1 = x1 * x1 + y1 * y1;
          const d2 = d1 - x2 * x2 - y2 * y2;
          const d3 = d1 - x3 * x3 - y3 * y3;
          const ab = (a3 * b2 - a2 * b3) * 2;
          circumcenters[j] = (b2 * d3 - b3 * d2) / ab;
          circumcenters[j + 1] = (a3 * d2 - a2 * d3) / ab;
        }

        // Compute exterior cell rays.
        for (let n = hull.length, p0, x0, y0, p1 = triangles[hull[n - 1]] * 2, x1 = points[p1], y1 = points[p1 + 1], i = 0; i < n; ++i) {
          p0 = p1, x0 = x1, y0 = y1, p1 = triangles[hull[i]] * 2, x1 = points[p1], y1 = points[p1 + 1];
          vectors[p0 * 2 + 2] = vectors[p1 * 2] = y0 - y1;
          vectors[p0 * 2 + 3] = vectors[p1 * 2 + 1] = x1 - x0;
        }
      }
      render(context) {
        const buffer = context == null ? context = new Path : undefined;
        const {delaunay: {halfedges, hull, triangles}, circumcenters, vectors} = this;
        for (let i = 0, n = halfedges.length; i < n; ++i) {
          const j = halfedges[i];
          if (j < i) continue;
          const ti = Math.floor(i / 3) * 2;
          const tj = Math.floor(j / 3) * 2;
          const xi = circumcenters[ti];
          const yi = circumcenters[ti + 1];
          const xj = circumcenters[tj];
          const yj = circumcenters[tj + 1];
          this._renderSegment(xi, yi, xj, yj, context);
        }
        for (let i = 0, n = hull.length; i < n; ++i) {
          const t = Math.floor(hull[i] / 3) * 2;
          const x = circumcenters[t];
          const y = circumcenters[t + 1];
          const v = triangles[hull[i]] * 4;
          const p = this._project(x, y, vectors[v + 2], vectors[v + 3]);
          if (p) this._renderSegment(x, y, p[0], p[1], context);
        }
        return buffer && buffer.value();
      }
      renderBounds(context) {
        const buffer = context == null ? context = new Path : undefined;
        context.rect(this.xmin, this.ymin, this.xmax - this.xmin, this.ymax - this.ymin);
        return buffer && buffer.value();
      }
      renderCell(i, context) {
        const buffer = context == null ? context = new Path : undefined;
        const points = this._clip(i);
        if (points === null) return;
        context.moveTo(points[0], points[1]);
        for (let i = 2, n = points.length; i < n; i += 2) {
          context.lineTo(points[i], points[i + 1]);
        }
        context.closePath();
        return buffer && buffer.value();
      }
      *cellPolygons() {
        const {delaunay: {points}} = this;
        for (let i = 0, n = points.length / 2; i < n; ++i) {
          yield this.cellPolygon(i);
        }
      }
      cellPolygon(i) {
        const polygon = new Polygon;
        this.renderCell(i, polygon);
        return polygon.value();
      }
      _renderSegment(x0, y0, x1, y1, context) {
        let S;
        const c0 = this._regioncode(x0, y0);
        const c1 = this._regioncode(x1, y1);
        if (c0 === 0 && c1 === 0) {
          context.moveTo(x0, y0);
          context.lineTo(x1, y1);
        } else if (S = this._clipSegment(x0, y0, x1, y1, c0, c1)) {
          context.moveTo(S[0], S[1]);
          context.lineTo(S[2], S[3]);
        }
      }
      contains(i, x, y) {
        if ((x = +x, x !== x) || (y = +y, y !== y)) return false;
        return this._step(i, x, y) === i;
      }
      find(x, y, i = 0) {
        if ((x = +x, x !== x) || (y = +y, y !== y)) return -1;
        let c;
        while ((c = this._step(i, x, y)) >= 0 && c !== i) i = c;
        return c;
      }
      _step(i, x, y) {
        const {delaunay: {points, triangles}, edges, index} = this;
        if (points.length === 0) return -1; // Empty triangulation.
        const j0 = index[i * 2];
        const j1 = index[i * 2 + 1];
        if (j0 === j1) return -1; // Coincident point.
        let c = i, k = edges[j0] * 3;
        let dc = (x - points[c * 2]) ** 2 + (y - points[c * 2 + 1]) ** 2;
        switch (i) { // Test previous point on triangle (for hull).
          case triangles[k]: k = triangles[k + 2]; break;
          case triangles[k + 1]: k = triangles[k]; break;
          case triangles[k + 2]: k = triangles[k + 1]; break;
        }
        let dk = (x - points[k * 2]) ** 2 + (y - points[k * 2 + 1]) ** 2;
        if (dk < dc) dc = dk, c = k;
        for (let j = j0; j < j1; ++j) {
          k = edges[j] * 3;
          switch (i) { // Test next point on triangle.
            case triangles[k]: k = triangles[k + 1]; break;
            case triangles[k + 1]: k = triangles[k + 2]; break;
            case triangles[k + 2]: k = triangles[k]; break;
          }
          dk = (x - points[k * 2]) ** 2 + (y - points[k * 2 + 1]) ** 2;
          if (dk < dc) dc = dk, c = k;
        }
        return c;
      }
      _cell(i) {
        const {index, edges, circumcenters} = this;
        const t0 = index[i * 2];
        const t1 = index[i * 2 + 1];
        if (t0 === t1) return null;
        const points = new Float64Array((t1 - t0) * 2);
        for (let t = t0, j = 0; t < t1; ++t, j += 2) {
          const ti = edges[t] * 2;
          points[j] = circumcenters[ti];
          points[j + 1] = circumcenters[ti + 1];
        }
        return points;
      }
      _clip(i) {
        const points = this._cell(i);
        if (points === null) return null;
        const {vectors: V} = this;
        const v = i * 4;
        return V[v] || V[v + 1]
            ? this._clipInfinite(i, points, V[v], V[v + 1], V[v + 2], V[v + 3])
            : this._clipFinite(i, points);
      }
      _clipFinite(i, points) {
        const n = points.length;
        let P = null;
        let x0, y0, x1 = points[n - 2], y1 = points[n - 1];
        let c0, c1 = this._regioncode(x1, y1);
        let e0, e1;
        for (let i = 0; i < n; i += 2) {
          x0 = x1, y0 = y1, x1 = points[i], y1 = points[i + 1];
          c0 = c1, c1 = this._regioncode(x1, y1);
          if (c0 === 0 && c1 === 0) {
            e0 = e1, e1 = 0;
            if (P) P.push(x1, y1);
            else P = [x1, y1];
          } else {
            let S, sx0, sy0, sx1, sy1;
            if (c0 === 0) {
              if ((S = this._clipSegment(x0, y0, x1, y1, c0, c1)) === null) continue;
              [sx0, sy0, sx1, sy1] = S;
            } else {
              if ((S = this._clipSegment(x1, y1, x0, y0, c1, c0)) === null) continue;
              [sx1, sy1, sx0, sy0] = S;
              e0 = e1, e1 = this._edgecode(sx0, sy0);
              if (e0 && e1) this._edge(i, e0, e1, P, P.length);
              if (P) P.push(sx0, sy0);
              else P = [sx0, sy0];
            }
            e0 = e1, e1 = this._edgecode(sx1, sy1);
            if (e0 && e1) this._edge(i, e0, e1, P, P.length);
            if (P) P.push(sx1, sy1);
            else P = [sx1, sy1];
          }
        }
        if (P) {
          e0 = e1, e1 = this._edgecode(P[0], P[1]);
          if (e0 && e1) this._edge(i, e0, e1, P, P.length);
        } else if (this.contains(i, (this.xmin + this.xmax) / 2, (this.ymin + this.ymax) / 2)) {
          return [this.xmax, this.ymin, this.xmax, this.ymax, this.xmin, this.ymax, this.xmin, this.ymin];
        }
        return P;
      }
      _clipSegment(x0, y0, x1, y1, c0, c1) {
        while (true) {
          if (c0 === 0 && c1 === 0) return [x0, y0, x1, y1];
          if (c0 & c1) return null;
          let x, y, c = c0 || c1;
          if (c & 0b1000) x = x0 + (x1 - x0) * (this.ymax - y0) / (y1 - y0), y = this.ymax;
          else if (c & 0b0100) x = x0 + (x1 - x0) * (this.ymin - y0) / (y1 - y0), y = this.ymin;
          else if (c & 0b0010) y = y0 + (y1 - y0) * (this.xmax - x0) / (x1 - x0), x = this.xmax;
          else y = y0 + (y1 - y0) * (this.xmin - x0) / (x1 - x0), x = this.xmin;
          if (c0) x0 = x, y0 = y, c0 = this._regioncode(x0, y0);
          else x1 = x, y1 = y, c1 = this._regioncode(x1, y1);
        }
      }
      _clipInfinite(i, points, vx0, vy0, vxn, vyn) {
        let P = Array.from(points), p;
        if (p = this._project(P[0], P[1], vx0, vy0)) P.unshift(p[0], p[1]);
        if (p = this._project(P[P.length - 2], P[P.length - 1], vxn, vyn)) P.push(p[0], p[1]);
        if (P = this._clipFinite(i, P)) {
          for (let j = 0, n = P.length, c0, c1 = this._edgecode(P[n - 2], P[n - 1]); j < n; j += 2) {
            c0 = c1, c1 = this._edgecode(P[j], P[j + 1]);
            if (c0 && c1) j = this._edge(i, c0, c1, P, j), n = P.length;
          }
        } else if (this.contains(i, (this.xmin + this.xmax) / 2, (this.ymin + this.ymax) / 2)) {
          P = [this.xmin, this.ymin, this.xmax, this.ymin, this.xmax, this.ymax, this.xmin, this.ymax];
        }
        return P;
      }
      _edge(i, e0, e1, P, j) {
        while (e0 !== e1) {
          let cx, cy;
          switch (e0) {
            case 0b0101: e0 = 0b0100; continue; // top-left
            case 0b0100: e0 = 0b0110, cx = this.xmax, cy = this.ymin; break; // top
            case 0b0110: e0 = 0b0010; continue; // top-right
            case 0b0010: e0 = 0b1010, cx = this.xmax, cy = this.ymax; break; // right
            case 0b1010: e0 = 0b1000; continue; // bottom-right
            case 0b1000: e0 = 0b1001, cx = this.xmin, cy = this.ymax; break; // bottom
            case 0b1001: e0 = 0b0001; continue; // bottom-left
            case 0b0001: e0 = 0b0101, cx = this.xmin, cy = this.ymin; break; // left
          }
          if (this.contains(i, cx, cy)) {
            P.splice(j, 0, cx, cy), j += 2;
          }
        }
        return j;
      }
      _project(x0, y0, vx, vy) {
        let t = Infinity, c, x, y;
        if (vy < 0) { // top
          if (y0 <= this.ymin) return null;
          if ((c = (this.ymin - y0) / vy) < t) y = this.ymin, x = x0 + (t = c) * vx;
        } else if (vy > 0) { // bottom
          if (y0 >= this.ymax) return null;
          if ((c = (this.ymax - y0) / vy) < t) y = this.ymax, x = x0 + (t = c) * vx;
        }
        if (vx > 0) { // right
          if (x0 >= this.xmax) return null;
          if ((c = (this.xmax - x0) / vx) < t) x = this.xmax, y = y0 + (t = c) * vy;
        } else if (vx < 0) { // left
          if (x0 <= this.xmin) return null;
          if ((c = (this.xmin - x0) / vx) < t) x = this.xmin, y = y0 + (t = c) * vy;
        }
        return [x, y];
      }
      _edgecode(x, y) {
        return (x === this.xmin ? 0b0001
            : x === this.xmax ? 0b0010 : 0b0000)
            | (y === this.ymin ? 0b0100
            : y === this.ymax ? 0b1000 : 0b0000);
      }
      _regioncode(x, y) {
        return (x < this.xmin ? 0b0001
            : x > this.xmax ? 0b0010 : 0b0000)
            | (y < this.ymin ? 0b0100
            : y > this.ymax ? 0b1000 : 0b0000);
      }
    }

    const tau = 2 * Math.PI;

    function pointX(p) {
      return p[0];
    }

    function pointY(p) {
      return p[1];
    }

    class Delaunay {
      constructor(points) {
        const {halfedges, hull, triangles} = new Delaunator(points);
        this.points = points;
        this.halfedges = halfedges;
        this.hull = Uint32Array.from(hullIterable(hull));
        this.triangles = triangles;
      }
      voronoi(bounds) {
        return new Voronoi(this, bounds);
      }
      render(context) {
        const buffer = context == null ? context = new Path : undefined;
        const {points, halfedges, triangles} = this;
        for (let i = 0, n = halfedges.length; i < n; ++i) {
          const j = halfedges[i];
          if (j < i) continue;
          const ti = triangles[i] * 2;
          const tj = triangles[j] * 2;
          context.moveTo(points[ti], points[ti + 1]);
          context.lineTo(points[tj], points[tj + 1]);
        }
        this.renderHull(context);
        return buffer && buffer.value();
      }
      renderPoints(context, r = 2) {
        const buffer = context == null ? context = new Path : undefined;
        const {points} = this;
        for (let i = 0, n = points.length; i < n; i += 2) {
          const x = points[i], y = points[i + 1];
          context.moveTo(x + r, y);
          context.arc(x, y, r, 0, tau);
        }
        return buffer && buffer.value();
      }
      renderHull(context) {
        const buffer = context == null ? context = new Path : undefined;
        const {points, hull, triangles} = this;
        const n = hull.length;
        let i0, i1 = triangles[hull[n - 1]] * 2;
        for (let i = 0; i < n; ++i) {
          i0 = i1, i1 = triangles[hull[i]] * 2;
          context.moveTo(points[i0], points[i0 + 1]);
          context.lineTo(points[i1], points[i1 + 1]);
        }
        return buffer && buffer.value();
      }
      hullPolygon() {
        const polygon = new Polygon;
        this.renderHull(polygon);
        return polygon.value();
      }
      renderTriangle(i, context) {
        const buffer = context == null ? context = new Path : undefined;
        const {points, triangles} = this;
        const t0 = triangles[i *= 3] * 2;
        const t1 = triangles[i + 1] * 2;
        const t2 = triangles[i + 2] * 2;
        context.moveTo(points[t0], points[t0 + 1]);
        context.lineTo(points[t1], points[t1 + 1]);
        context.lineTo(points[t2], points[t2 + 1]);
        context.closePath();
        return buffer && buffer.value();
      }
      *trianglePolygons() {
        const {triangles} = this;
        for (let i = 0, n = triangles.length / 3; i < n; ++i) {
          yield this.trianglePolygon(i);
        }
      }
      trianglePolygon(i) {
        const polygon = new Polygon;
        this.renderTriangle(i, polygon);
        return polygon.value();
      }
    }

    Delaunay.from = function(points, fx = pointX, fy = pointY, that) {
      return new Delaunay("length" in points
          ? flatArray(points, fx, fy, that)
          : Float64Array.from(flatIterable(points, fx, fy, that)));
    };

    function flatArray(points, fx, fy, that) {
      const n = points.length;
      const array = new Float64Array(n * 2);
      for (let i = 0; i < n; ++i) {
        const p = points[i];
        array[i * 2] = fx.call(that, p, i, points);
        array[i * 2 + 1] = fy.call(that, p, i, points);
      }
      return array;
    }

    function* flatIterable(points, fx, fy, that) {
      let i = 0;
      for (const p of points) {
        yield fx.call(that, p, i, points);
        yield fy.call(that, p, i, points);
        ++i;
      }
    }

    function* hullIterable(hull) {
      let node = hull;
      do yield node.t;
      while ((node = node.next) !== hull);
    }

    exports.Delaunay = Delaunay;
    exports.Voronoi = Voronoi;

    Object.defineProperty(exports, '__esModule', { value: true });

})));
